#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Fri 13 May 2011 10:18:26 AM CST

import sys
import time
import uuid

try:
    from nas.system import sqlite, HOMEDB
    from nas.system import getrealpath
    from nas.system import getunicode, getlist
    from nas.system import NASError
except ImportError:
    from system import sqlite, HOMEDB
    from system import getrealpath
    from system import getunicode, getlist
    from system import NASError

def _getcursor(con):
    cur = con.cursor()
    cur.execute(''' CREATE TABLE
                    IF NOT EXISTS
                    publish (
                        id INTEGER PRIMARY KEY,
                        path TEXT NOT NULL,
                        owner TEXT NOT NULL,
                        vpath TEXT NOT NULL,
                        guest TEXT NOT NULL,
                        expire REAL NOT NULL,
                        rw INTEGER NOT NULL,
                        uuid TEXT NOT NULL,
                        other TEXT,
                        UNIQUE(path, guest),
                        UNIQUE(id, uuid)
                    );
                ''')
    cur.execute(''' DELETE FROM publish WHERE expire < ?
                ''', (time.time(),))
    con.commit()
    return cur

def publish(path, guest, expire, rw, login=None, **kw):
    '''
    path: path
    guest: guest, can be multi
    expire: expire date, YYYYMMDD, can be multi
    rw: yes | no , can be multi

    Note: if expire is invalid, the record will be deleted
    '''

    realpath = getrealpath(path, login, None)
    if not realpath:
        raise NASError('Invalid Path')

    def mktime(x):
        try:
            date = time.mktime(time.strptime(x, '%Y%m%d'))
        except:
            date = 0
        return date

    def getrw(x):
        if x == 'yes':
            return 1
        else:
            return 0

    s_guest = getlist(guest)
    s_expire = [mktime(x) for x in getlist(expire)]
    s_rw = [getrw(x) for x in getlist(rw)]

    con = sqlite.connect(HOMEDB)
    cur = _getcursor(con)

    db_path = getunicode(realpath)
    db_owner = getunicode(login)
    db_vpath = getunicode(path)
    for db_guest, db_expire, db_rw in zip(s_guest, s_expire, s_rw):
        if not db_guest:
            continue

        if not db_expire:
            cur.execute(''' DELETE FROM publish WHERE path = ? AND guest = ?
                        ''', (db_path, db_guest))
            continue

        cur.execute(''' UPDATE publish SET expire = ?, rw = ?
                        WHERE path = ? AND guest = ?
                    ''', (db_expire, db_rw, db_path, db_guest))
        if cur.rowcount:
            continue

        db_uuid = unicode(uuid.uuid4())
        try:
            cur.execute(''' INSERT INTO publish (path, owner, vpath, guest, expire, rw, uuid)
                            VALUES (?, ?, ?, ?, ?, ?, ?)
                        ''', (db_path, db_owner, db_vpath, db_guest, db_expire, db_rw, db_uuid))
        except sqlite.IntegrityError:
            continue

    con.commit()
    con.close()

    record = resolvepath(db_path)
    return {'publish': {'path': path,
                        'record': record}}


def unpublish(path, login=None, **kw):
    db_vpath = getunicode(path)
    con = sqlite.connect(HOMEDB)
    cur = _getcursor(con)
    cur.execute(''' DELETE FROM publish WHERE vpath = ?
                ''', (db_vpath,))
    con.commit()
    con.close()
    return {'unpublish' : {'path' : path}}


def resolvepath(path):
    '''
    resolve a path to 'guest, expire, rw, uuid'
    Note: only for internal use now
    '''

    assert isinstance(path, unicode)

    con = sqlite.connect(HOMEDB)
    cur = _getcursor(con)

    cur.execute(''' SELECT guest, expire, rw, uuid FROM publish
                    WHERE path = ?
                ''', (path,))

    result = []
    for guest, expire, rw, uuid in cur.fetchall():
        result.append({'guest': guest.encode('utf8'),
                       'expire': time.strftime('%Y%m%d', time.localtime(expire)),
                       'rw': rw and 'yes' or 'no',
                       'uuid': uuid.encode('utf8'),
                       })

    con.close()
    return result

def resolveowner(login, **kw):
    '''
    resolve the path for current login user
    '''

    db_owner = getunicode(login)

    con = sqlite.connect(HOMEDB)
    cur = _getcursor(con)

    cur.execute(''' SELECT vpath, guest, expire, rw, uuid FROM publish
                    WHERE owner = ?
                ''', (db_owner,))

    result = []
    for vpath, guest, expire, rw, uuid in cur.fetchall():
        realpath = getrealpath(vpath.encode('utf8'), login, None)
        if not realpath:
            continue
        result.append({'path': vpath.encode('utf8'),
                       'guest': guest.encode('utf8'),
                       'expire': time.strftime('%Y%m%d', time.localtime(expire)),
                       'rw': rw and 'yes' or 'no',
                       'uuid': uuid.encode('utf8'),
                       })

    con.close()
    return {'publish': {'owner': db_owner.encode('utf8'),
                        'record': result}}

def resolveuuid(uuid):
    '''
    resolve a uuid to path, guest, expire, rw

    returns: path, guest, expire(float), rw(1|0)
    '''

    db_uuid = unicode(uuid)

    con = sqlite.connect(HOMEDB)
    cur = _getcursor(con)

    cur.execute(''' SELECT path, guest, expire, rw FROM publish
                    WHERE uuid = ?
                ''', (db_uuid,))

    try:
        path, guest, expire, rw = cur.fetchone()
    except (TypeError, ValueError):
        path = guest = expire = rw = None

    con.close()

    if path:
        path = path.encode('utf8')

    return (path, guest, expire, rw)

if __name__ == '__main__':
    query = {
            'path': '/home/admin',
            'guest': 'admin',
            'expire': '20110930',
            'rw': 'no',
            }
    # result = publish(**query)
    print resolveuuid('b173da85-9130-4c3a-b785-7ff6101beaec')

# vim: set sta sw=4 et:
