#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Mon 07 Jun 2010 02:14:13 PM CST

import parted
from _ped import DiskLabelException

import re
import os
import sys
import errno
import fcntl
import time

from system import NASError
from system import POPEN
from system import makesure_nasdir
from system import echo
from system import backup, unbackup
from system import HOME

import glob
import itertools

from service import stopservice
from mlnet import setmlnet
from daap import setdaap
from dlna import setdlna
from samba import setsmb
from samba import addsmbdev
from samba import checksmbpath
from nfs import setexports
from ftpd import setftp
from smtp import _notify
from afpd import addafpdev
from afpd import rmafpdev
FSTAB = '/etc/fstab'
MDADMCONF = '/etc/mdadm.conf'

FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC | os.O_SYNC
FSTABCONTENT = '''
ubi0    /               ubifs   defaults        1 1
devpts  /dev/pts        devpts  gid=5,mode=620  0 0
proc    /proc           proc    defaults        0 0
sysfs   /sys            sysfs   defaults        0 0
none    /var/tmp        tmpfs   defaults        0 0
sunrpc  /var/lib/nfs/rpc_pipefs rpc_pipefs      defaults 0 0
# mount data partition to /home
'''

if not os.path.exists('/proc/mdstat'):
    (error, result) = POPEN('modprobe md_mod')
    result.close()

PARTITION_TYPE = {0: 'primary',
                  1: 'logic',
                  2: 'extended',
                  3: 'free',
                  4: 'metadata',
                  5: 'protected',
                  }

class CommitError(NASError):
    def __init__(self, *kw):
        self.reason = kw[0]
        self.detail = kw[1]
        self.args = kw

disktype = lambda x: x > 4294967295 and 'gpt' or 'msdos'

def getuuid(path):
    try:
        cmd = "blkid %s" % path
        pipe = os.popen(cmd)
        content = re.findall("(?<=UUID=\").*?(?=\")", pipe.read())[0]
        pipe.close()
    except:
        return path
    return content

def _getudevinfo(path):
    error, status = POPEN(['udevadm', 'info', '--query=property', '--name=%s' % path])
    json = []
    while error == 0:
        line = status.readline()
        if not line:
            break
        a, b, c = line.partition('=')
        json.append( '\"' + a.strip() + '\":\"' + c.strip() + '\"' )
    status.close()
    s= "{" + ",".join(json)+ "}"
    print s
    return s


def _getmountpoints():
    file = open('/proc/mounts')
    mountpoints = {}
    while True:
        line = file.readline()
        if not line:
            break
        device, directory = line.split()[:2]
        if device[:4] == '/dev':
            try:
                os.stat(device)
            except os.error:
                error, status = POPEN(['umount', '-r', directory])
                status.close()
                continue
        if device in mountpoints and directory not in mountpoints[device]:
            mountpoints[device].append(directory)
        else:
            mountpoints[device] = [directory]
    file.close()
    return mountpoints

def _getmdstat(device, state=True):
    path = '/proc/mdstat'
    if not os.path.exists(path):
        return (None, None, None, None)

    realpath = os.path.realpath(device)
    shortname = os.path.split(realpath)[1]

    file = open(path)
    while True:
        line = file.readline()
        if not line:
            break
        if not line.startswith(shortname):
            continue
        lists = line.split(':')[1].strip().split()
        devices = []
        for x in lists[2:]:
            path = os.path.join('/dev', x.partition('[')[0])
            stat = ''
            if state:
                if '(F)' in x:
                    stat = ' (faulty)'
                elif '(S)' in x:
                    stat = ' (spare)'
            devices.append('%s%s' % (path, stat))
        sync = ''
        line = file.readline()
        while line and not line.startswith('md'):
            line = file.readline()
            if 'speed' in line:
                sync = line.strip()
                break
        file.close()
        return (lists[0], lists[1], devices, sync)
    file.close()
    return (None, None, None, None)

def getdevices(**kw):
    dxmls = []

    for path in glob.glob('/dev/[h|s]d[a-z]'):
        dxml = getdevice(path)
        if dxml:
            dxmls.append(dxml)

    for path in glob.glob('/dev/md[0-9]*'):
        if 'p' in path:
            continue
        dxml = getdevice(path)
        if dxml:
            dxmls.append(dxml)

    return {'device': dxmls}

def getdevice(path=None, **kw):
    if path is None:
        raise NASError('Bad Request')

    dxml = {}

    if isinstance(path, basestring):
        path = os.path.realpath(path)
        if path.startswith('/dev/sd'):
            if len(path) != len('/dev/sda'):
                return dxml
        elif path.startswith('/dev/md'):
            if 'p' in path:
                return dxml
        else:
            return dxml
        try:
            device = parted.Device(path)
        except parted.DeviceException:
            return dxml
        except parted.IOException:
            return dxml
    else:
        device = path

    if device.length == 0:
        return dxml

    # /dev/mdXpY -> /dev/mdX
    raid = []
    if device.path.startswith('/dev/md'):
        # /dev/mdXpY
        if 'p' in device.path:
            return dxml
        raid = {}
        (state, level, devices, sync) = _getmdstat(device.path)
        if state and level and devices:
            raid['level'] = level
            raid['state'] = state
            if sync:
                raid['sync'] = sync
                percent = re.findall('([0-9.]+%)', sync)
                if percent:
                    raid['syncpercent'] = percent[0]
            raid['raiddevice'] = devices
            argv = ['--misc', '--detail', device.path]
            (error, status) = _mdadm(argv)
            raid['detail'] = status

    geometry = device.biosGeometry
    dxml['path'] = device.path
    dxml['cylinder'] = '%s' % geometry[0]
    dxml['head'] = '%s' % geometry[1]
    dxml['sector'] = '%s' % geometry[2]
    dxml['length'] = '%s' % device.length
    dxml['sectorsize'] = '%s' % device.sectorSize
    member = _getmdpath(device.path, True)
    if member:
        dxml['raidmember'] = member
    if raid:
        dxml['raid'] = raid

    if device.path[:7] == '/dev/sd':
        dxml['model'] = 'UNKNOWN'
        error, status = POPEN(['udevadm', 'info', '--query=property', '--name=%s' % device.path])
        while error == 0:
            line = status.readline()
            if not line:
                break
            a, b, c = line.partition('=')
            if a == 'ID_SERIAL' and b == '=':
                dxml['serial'] = c.strip()
            elif a == 'ID_MODEL' and b == '=':
                dxml['model'] = c.strip()
            elif a == 'ID_BUS' and b == '=':
                dxml['connectiontype'] = c.strip()
        status.close()
    else:
        dxml['model'] = 'Linux Soft RAID'

    try:
        disk = parted.Disk(device)
    except DiskLabelException:
        return dxml
    except parted.DiskException:
        return dxml
    dxml['type'] = '%s' % disk.type

    mountpoints = _getmountpoints()

    for partition in disk.partitions:
        geometry = partition.geometry
        pxml = {}
        # loop is very special
        if disk.type == 'loop':
            realpath = device.path
        else:
            realpath = partition.path

        pxml['path'] = '%s' % realpath

        member = _getmdpath(realpath, True)
        if member:
            pxml['raidmember'] = member
        elif partition.type in [0, 1]:
            label = _getlabel(realpath)
            fstype = _getfstype(realpath)
            if not fstype and partition.fileSystem:
                fstype = partition.fileSystem.type
            if label:
                pxml['label'] = '%s' % label
            if fstype != 'linux_raid_member':
                pxml['fs'] = fstype                
            if fstype == 'swap' and realpath in _getswap():
                pxml['mountpoint'] = realpath
            elif realpath in mountpoints:
                pxml['mountpoint'] = mountpoints[realpath]

                vfs = os.statvfs(mountpoints[realpath][0])
                pxml['blocks'] = str(vfs.f_blocks * vfs.f_bsize)
                pxml['used'] = str((vfs.f_blocks - vfs.f_bavail) * vfs.f_bsize)
                pxml['available'] = str(vfs.f_bavail * vfs.f_bsize)
                pxml['usedpercent'] = str(100 - 100 * vfs.f_bavail / vfs.f_blocks)

        pxml['start'] = '%s' % geometry.start
        pxml['end'] = '%s' % geometry.end
        pxml['length'] = '%s' % geometry.length
        pxml['type'] = '%s' % PARTITION_TYPE[partition.type]
        pxml['uuid'] =  _blkid(partition, 'UUID') #UUIDϢ
        if partition.getFlagsAsString():
            pxml['flags'] = '%s' % partition.getFlagsAsString()

        if 'partition' not in dxml:
            dxml['partition'] = [pxml]
        else:
            dxml['partition'].append(pxml)

        # has another partition?
        if disk.type == 'loop':
            break

    return dxml

def _checkdevice(path):
    if path.startswith('/dev/sd'):
        if len(path) != len('/dev/sda'):
            raise NASError('Invalid Device', path)
    elif path.startswith('/dev/md'):
        if 'p' in path:
            raise NASError('Invalid Device', path)
    else:
        raise NASError('Invalid Device', path)

def _checkpartition(path):
    if path.startswith('/dev/sd'):
        if len(path) == len('/dev/sda'):
            raise NASError('Invalid Partition', path)
        return path[:8]
    elif path.startswith('/dev/md'):
        if 'p' not in path:
            raise NASError('Invalid Partition', path)
        return path.split('p')[0]
    else:
        raise NASError('Invalid Partition', path)

def _checkmounted(path):
    mountpoints = _getmountpoints()
    mounted = []
    for mountpoint in mountpoints:
        if mountpoint.startswith(path):
            mounted.append(mountpoint)

    if mounted:
        raise NASError('Mounted Device', ', '.join(mounted))

# todo: umount path before action
def deleteallpartition(path=None, inter=False, **kw):
    if path is None:
        raise NASError('Bad Request')

    path = os.path.realpath(str(path))

    _checkdevice(path)
    _checkmounted(path)

    for x in _getswap():
        if x.startswith(path):
            nil, status = POPEN(['swapoff', x])
            status.close()

    try:
        device = parted.Device(path)
    except parted.DeviceException as error:
        raise NASError('Invalid Device', error)
    except parted.IOException as error:
        raise NASError('Invalid Device', error)

    if device.busy:
        raise NASError('Busy Device', path)

    try:
        # todo: gpt
        disk = parted.freshDisk(device, disktype(device.length))
        time.sleep(3)
        disk.deleteAllPartitions()
        time.sleep(3)
        disk.commit()
    except parted.PartitionException as error:
        print >> sys.stderr, error
        raise CommitError('Delete All Partitions Commit Error, Please Reboot', str(error))
    except parted.DiskException as error:
        print >> sys.stderr, error
        raise CommitError('Delete All Partitions Commit Error, Please Reboot', str(error))

    if inter:
        return True

    # return getdevice(path)
    return getdevices()

def deletepartition(path=None, **kw):
    '''
    extend partitions

    normal action:
    1 2 3 4
          5 6 7
       || ------> del 6
    1 2 3 4
          5   6

    now action:
    1 2 3 4
          5 6 7
       || ------> del 6
    1 2 3 4
          . . .
    '''

    if path is None:
        raise NASError('Bad Request')

    path = os.path.realpath(path)
    devicepath = _checkpartition(path)
    _checkmounted(path)

    try:
        device = parted.Device(devicepath)
    except parted.DeviceException as error:
        raise NASError('Invalid Partition', error)
    except parted.IOException as error:
        raise NASError('Invalid Partition', error)

    try:
        disk = parted.Disk(device)
    except DiskLabelException:
        disk = parted.freshDisk(device, disktype(device.length))
    except parted.DiskException as error:
        raise NASError('Invalid Disk', error)

    for partition in disk.partitions:
        if partition.path == path:
            if partition.busy:
                raise NASError('Busy Partition', partition.path)
            disk.deletePartition(partition)
            try:
                disk.commit()
            except parted.DiskException as error:
                print >> sys.stderr, error
                raise CommitError('Delete Partition Commit Error, Please Reboot', str(error))
    #return getdevice(device.path)
    return getdevices()

def addpartition(path=None, start=None, length=None, flag=None, fs=None, inter=False, **kw):
    if path is None:
        raise NASError('Bad Request')

    path = os.path.realpath(path)
    _checkdevice(path)
    _checkmounted(path)

    try:
        device = parted.Device(path)
    except parted.DeviceException as error:
        raise NASError('Invalid Device', error)
    except parted.IOException as error:
        raise NASError('Invalid Device', error)

    #if device.busy:
    #    raise NASError('Busy Device', device.path)

    try:
        disk = parted.Disk(device)
    except DiskLabelException:
        disk = parted.freshDisk(device, disktype(device.length))
    except parted.DiskException as error:
        raise NASError('Invalid Disk', error)

    cylinder, head, sector = device.biosGeometry

    #if start < sector:
    #    start = sector
    if disk.type == 'loop':
        minstart = 0
    else:
        minstart = sector

    len_corrector = 0
    if start is not None and str(start).isdigit():
        start = int(start)
        if start < minstart:
            len_corrector = minstart - start
            start = minstart
    else:
        start = minstart

    maxlength = cylinder * head * sector - start
    reallength = maxlength - (maxlength % (4 * 1024)) # length must be the 4KB times
    if reallength < sector:
        raise NASError('Invalid Geometry')

    if length is not None and str(length).isdigit():
        length = int(length) - len_corrector
        if length > reallength:
            length = reallength
    else:
        length = reallength

    if disk.type != 'gpt' and length > 4294967295:
        disk = parted.freshDisk(device, 'gpt')

    geometry = parted.Geometry(device, start, length)
    constraint = device.getConstraint()

    for partition in disk.partitions:
        if partition.type == parted.PARTITION_EXTENDED:
            continue
#        if partition.geometry.containsSector(geometry.start) or \
#                partition.geometry.containsSector(geometry.end):
#            raise NASError('Invalid Geometry', (partition.geometry.start, partition.geometry.end))

    # default is primary
    ptype = parted.PARTITION_NORMAL
    if kw.get('type') is not None:
        for _type in PARTITION_TYPE:
            if kw.get('type') == PARTITION_TYPE[_type]:
                ptype = _type
                break

    pfs = None
    if hasattr(fs, 'type') and fs.type in parted.fileSystemType:
        pfs = parted.fileSystemType[fs.type]
    elif fs is not None and str(fs) in parted.fileSystemType:
        pfs = parted.fileSystemType[str(fs)]

    pflag = None
    if flag is not None:
        if str(flag) == 'swap':
            pflag = parted.PARTITION_SWAP
        elif str(flag) == 'raid':
            pflag = parted.PARTITION_RAID
    if ptype == parted.PARTITION_EXTENDED:
        pflag = parted.PARTITION_LBA
    elif hasattr(pfs, 'type') and pfs.type == 'swap':
        pflag = parted.PARTITION_SWAP

    if not hasattr(pfs, 'type'):
        pfs = None

    partition = parted.Partition(disk, ptype, pfs, geometry)

    print >> sys.stderr, '[debug]', ptype, pfs, geometry.start, geometry.end
    print >> sys.stderr, '[debug]', partition.path, partition.geometry.start, partition.geometry.end
    print >> sys.stderr, '[debug]', partition

    try:
        disk.addPartition(partition, constraint)
    except parted.PartitionException as error:
        raise NASError('Parted Error', error)

    newpath = partition.path

    try:
        disk.commit()
    except parted.DiskException as error:
        print >> sys.stderr, error
        raise CommitError('Add Partition Commit Error, Please Reboot', str(error))

    # how to reload ?
    try:
        disk = parted.Disk(device)
    except DiskLabelException:
        disk = parted.freshDisk(device, disktype(device.length))
    except parted.DiskException as error:
        raise NASError('Invalid Disk', error)

    for partition in disk.partitions:
        if newpath != partition.path:
            continue

        if pflag:
            try:
                partition.setFlag(pflag)
            except parted.PartitionException:
                pass

        filesystem = partition.fileSystem
        if not filesystem:
            break

        if pflag in [parted.PARTITION_LBA, parted.PARTITION_RAID] or pfs is None or inter:
            print >> sys.stderr, 'clobber filesystem for %s' % partition.path
            partition.fileSystem.clobber()
            filesystem = None

    try:
        disk.commitToDevice()
    except parted.DiskException as error:
        print >> sys.stderr, error
        raise NASError('CommitToDevice Error, Please Reboot', str(error))

    if inter:
        print >> sys.stderr, 'inter request, return', partition.path
        return (partition.path, partition.geometry)
    # return getdevice(path)
    return getdevices()

def umount(mountpoint=None, static='no', **kw):
    if mountpoint is None:
        raise NASError('Bad Request')
    if mountpoint == '':
        raise NASError('Bad Request')

    if isinstance(mountpoint, list):
        mountpoints = [str(x) for x in mountpoint]
    else:
        mountpoints = [str(mountpoint)]

    for mountpoint in mountpoints:
        mountpoint = os.path.realpath(mountpoint)
        if mountpoint == '/home':
            stopservice('nfs')
            stopservice('samba')
            stopservice('mt-daapd')
            stopservice('minidlna')
            stopservice('mldonkey')
            stopservice('vsftpd')
            try:
                import nas.thunder
                stopservice('thunder')
            except:
                pass
        if mountpoint in _getswap():
            error, status = POPEN(['swapoff', mountpoint])
        elif mountpoint.startswith('/dev/'):
            raise NASError('Umount Error', mountpoint)
        else:
            error, status = POPEN(['umount', '-l', mountpoint])
            error, status = POPEN(['nas-mountpoint', 'remove', mountpoint])

        result = status.read()
        status.close()

        static = str(static)
        if static == 'yes':
            _removefstab(mountpoint)

        if error != 0:
            raise NASError('System Error', result)
    checksmbpath(mountpoint)
    rmafpdev(mountpoint)
    return {'umount': 'umount %s ok' % ', '.join(mountpoints)}

def write(pipe, data):
    if os.write(pipe, data) != len(data):
        return 1
    return 0

def _removefstab(path):
    path = os.path.realpath(path)
    uuid = getuuid(path)
    backup(FSTAB)

    try:
        os.stat(FSTAB)
    except os.error:
        src = None
    else:
        src = open(FSTAB, 'r')

    dst = os.open('%s.lock' % FSTAB, FLAGS, 0644)
    fcntl.lockf(dst, fcntl.LOCK_EX)

    error = 0
    while error == 0:
        if src is None:
            error |= write(dst, FSTABCONTENT)
            break
        line = src.readline()
        if not line:
            break
        line = line.lstrip()
        if line[:1] == '#':
            error |= write(dst, line)
            continue
        fstab = line.split()[:6]
        if len(fstab) != 6:
            error |= write(dst, line)
            continue
        if os.path.realpath(fstab[1]) == path or uuid in fstab[0]:
            continue
        else:
            error |= write(dst, line)

    os.fsync(dst)
    fcntl.lockf(dst, fcntl.LOCK_UN)
    os.close(dst)
    if src is not None:
        src.close()

    if error:
        raise NASError('write data error', FSTAB)

    os.rename('%s.lock' % FSTAB, FSTAB)
    unbackup(FSTAB)

    return True

def _writefstab(path, mountpoint, mounttype, mountoption):
    file = open('/proc/mounts')
    while True:
        record = file.readline()
        if not record:
            if mountpoint == 'none':
                break
            file.close()
            return False
        if record.split()[0] == path:
            break
    if mounttype == 'relatime':
        mounttype = 'defaults'
    if mounttype != 'auto':
        record = 'UUID=%s %s %s %s 0 0' % (getuuid(path), mountpoint, mounttype, mountoption)
    file.close()

    backup(FSTAB)

    try:
        os.stat(FSTAB)
    except os.error:
        src = None
    else:
        src = open(FSTAB, 'r')

    dst = os.open('%s.lock' % FSTAB, FLAGS, 0644)
    fcntl.lockf(dst, fcntl.LOCK_EX)

    error = 0
    while error == 0:
        if src is None:
            error |= write(dst, FSTABCONTENT)
            break

        line = src.readline()
        if not line:
            break
        line = line.lstrip()
        if line[:1] == '#':
            error |= write(dst, line)
            continue
        fstab = line.split()[:6]
        if len(fstab) != 6:
            error |= write(dst, line)
            continue
        if record and os.path.realpath(record.split()[0]) == os.path.realpath(fstab[0]):
            error |= write(dst, '%s\n' % record.strip())
            record = ''
        else:
            error |= write(dst, line)
    if record and error == 0:
        error |= write(dst, '%s\n' % record.strip())

    os.fsync(dst)
    fcntl.lockf(dst, fcntl.LOCK_UN)
    os.close(dst)
    if src is not None:
        src.close()

    if error:
        raise NASError('write data error', FSTAB)

    os.rename('%s.lock' % FSTAB, FSTAB)
    unbackup(FSTAB)
    return True

def _mountswap(path, static):
    argv = ['swapon', path]
    error, result = POPEN(argv)
    status = result.read()
    result.close()
    if error != 0:
        raise NASError('SystemError', status)
    if str(static) == 'yes':
        _writefstab(path, 'none', 'swap', 'defaults')

def mount(path=None, mountpoint=None, static='no', **kw):
    from spindown import _spindown
    _spindown(path)

    if str(path) and _getfstype(str(path)) == 'swap':
        _mountswap(path, static)
        return {'path': path, 'mountpoint': 'swap'}

    if mountpoint == 'swap':
        _mountswap(path, static)
        return {'path': path, 'mountpoint': 'swap'}

    if path is None or mountpoint is None:
        raise NASError('Bad Request')

    path = os.path.realpath(str(path))
    mountpoint = str(mountpoint)

    for escape in ['\040', '\011', '\012', '\134']:
        if escape in mountpoint:
            raise NASError('Unsupported Directory %s' % mountpoint)

    if not path.startswith('/dev'):
        raise NASError('Invalid Device', path)
    if '/' in mountpoint:
        raise NASError('Invalid Mountpoint', mountpoint)

    if mountpoint == 'home':
        mountpoint = '/home'
    else:
        mountpoint = os.path.realpath(os.path.join('/media', mountpoint))

    if not os.path.isdir(mountpoint):
        try:
            os.makedirs(mountpoint)
            os.system('chattr +i %s' % mountpoint)
        except os.error:
            pass

    fstype = _getfstype(path)
    if fstype == 'ntfs':
        mounttype = 'ntfs-3g'
    elif fstype:
        mounttype = fstype
    else:
        mounttype = 'auto'
    if fstype in ['ntfs', 'vfat']:
        mountoption = 'utf8,dmask=000,fmask=111'
    elif fstype == 'xfs':
        mountoption = 'rw'
    else:
        mountoption = 'relatime'

    (error, result) = POPEN(['mount', '-t', mounttype, '-o', mountoption, path, mountpoint])
    status = result.read()
    result.close()

    if error != 0:
        if fstype == 'xfs':
            argv = ['xfs_repair', '-L', path]
            nil, status = POPEN(argv)
            status.close()
            if mountpoint == '/home':
                mountpoint = 'home'
            return mount(path, mountpoint, static)

        raise NASError('System Error', status)
    if mountpoint[:6] == '/media':
        addsmbdev(mountpoint)
        addafpdev(mountpoint)
        
    static = str(static)
    if static == 'yes':
        _writefstab(path, mountpoint, mounttype, mountoption)

    makesure_nasdir()

    service = {}
    if mountpoint == '/home' and static == 'yes':
#        service['nfs'] = setexports(enable='yes')
        service['samba'] = setsmb(enable='yes')
        service['daap'] = setdaap(enable='yes')
        service['dlna'] = setdlna(enable='yes')
        service['ftps'] = setftp(enable='yes')
        try:
            import nas.thunder
            service['thunder'] = nas.thunder.setthunder(enable='yes')
        except:
            service['mldonkey'] = setmlnet(enable='yes')
            pass

    return {'path': path, 'mountpoint': mountpoint, 'service': service}

def _fixflag(path, flag, fs=None):
    # /dev/sdaX, /dev/mdXpY
    loop = False
    if path.startswith('/dev/md') and 'p' not in path:
        loop = True
    elif path == path.rstrip('0123456789'):
        loop = True

    if loop:
        try:
            device = parted.Device(path)
        except parted.DeviceException as error:
            print >> sys.stderr, '[exception]', error
            return False
        except parted.IOException as error:
            print >> sys.stderr, '[exception]', error
            return False

        if not device.path.startswith('/dev/md'):
            argv = ['--zero-superblock', device.path]
            _mdadm(argv)

        disk = parted.freshDisk(device, 'loop')

        try:
            disk.commitToDevice()
        except parted.DiskException as error:
            print >> sys.stderr, '[exception]', error
            return False
        return True

    if path.startswith('/dev/md'):
        devicepath = path.partition('p')[0]
    else:
        devicepath = path.rstrip('0123456789')

    try:
        device = parted.Device(devicepath)
    except parted.DeviceException as error:
        print >>sys.stderr, '[exception]', error
        return False
    except parted.IOException as error:
        print >>sys.stderr, '[exception]', error
        return False

    try:
        disk = parted.Disk(device)
    except DiskLabelException:
        disk = parted.freshDisk(device, disktype(device.length))
    except parted.DiskException as error:
        raise NASError('Invalid Disk', error)

    if disk.type == 'loop':
        disk = parted.freshDisk(device, disktype(device.length))

    for partition in disk.partitions:
        if partition.path == path:
            break
    else:
        return False

    for Flag in parted.partitionFlag:
        if Flag == parted.PARTITION_LBA:
            continue
        elif Flag == flag:
            try:
                partition.setFlag(Flag)
            except parted.PartitionException:
                pass
        elif partition.getFlag(Flag):
            try:
                partition.unsetFlag(Flag)
            except parted.PartitionException:
                pass

    if flag == parted.PARTITION_RAID:
        if partition.fileSystem:
            partition.fileSystem.clobber()

    if fs is not None and fs in parted.fileSystemType:
        partition.system = parted.fileSystemType[fs]

    try:
        disk.commitToDevice()
    except parted.DiskException as error:
        return False

    # parted / pyparted bug
    if fs in ['hfs', 'hfs+', 'hfsplus']:
        if len(disk.partitions) == 1:
            command = 't\naf\nw\n'
        else:
            command = 't\n%d\naf\nw\n' % partition.number
        #os.system('''/bin/echo 'echo -n "%s" | fdisk %s' ''' % (command, device.path))
        (error, stdin) = POPEN('echo -n "%s"' % command)
        (error, fdisk) = POPEN('fdisk %s' % device.path, stdin)
        stdin.close()
        fdisk.close()

    argv = ['--zero-superblock', device.path]
    _mdadm(argv)

    argv = ['--zero-superblock', partition.path]
    _mdadm(argv)

    return True

# should use class, but i'm so lazy...
MKFS = {'fat16': 'mkdosfs -F 16 -I',
        'fat32': 'mkdosfs -F 32 -I',
        'ntfs': 'mkntfs -F -f',
        'ext2': 'mkfs.ext2 -F',
        'ext3': 'mkfs.ext3 -F',
        'ext4': 'mkfs.ext4 -F',
        'reiserfs': 'mkreiserfs -f -q',
        'jfs': 'mkfs.jfs -q',
        'xfs': 'mkfs.xfs -f',
        'swap': 'mkswap -f',
        'hfs': 'mkfs.hfsplus -h ',
        'hfs+': 'mkfs.hfsplus',
        'hfsplus': 'mkfs.hfsplus',
        }
MKFSOPT = {'fat16': '-n',
           'fat32': '-n',
           'ntfs': '-L',
           'ext2': '-L',
           'ext3': '-L',
           'ext4': '-L',
           'reiserfs': '-l',
           'jfs': '-L',
           'xfs': '-L',
           'swap': '-L',
           'hfs': '-v',
           'hfs+': '-v',
           'hfsplus': '-v',
           }

def getmkfs(path, **kw):
    path = str(path)
    path = os.path.realpath(path)
    lockfile = '/var/run/%s.lock' % os.path.split(path)[1]
    if not os.path.exists(lockfile):
        raise NASError('Bad Request')

    print 'getmkfs'

    lockfp = open(lockfile, 'a+')
    try:
        fcntl.lockf(lockfp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError as error:
        if error.errno not in [errno.EACCES, errno.EAGAIN]:
            lockfp.close()
            raise NASError('System Error', error.strerror)
    else:
        fcntl.lockf(lockfp, fcntl.LOCK_UN)
        # os.unlink(lockfp)

    path = lockfp.readline().strip()
    fs = lockfp.readline().strip()
    label = lockfp.readline().strip()
    force = lockfp.readline().strip()
    start = lockfp.readline().strip()

    detail = ''
    retcode = ''
    while True:
        line = lockfp.readline()
        if not line:
            break
        if line[:9] == 'retcode: ':
            retcode = line[9:]
            break
        else:
            detail += line

    lockfp.close()

    detail = echo(detail.strip())
    return {'getmkfs': {'path': path,
                        'fs': fs,
                        'label': label,
                        'force': force,
                        'start': start,
                        'detail': detail,
                        'retcode': retcode}}

def mkfs(path=None, fs=None, label='', force='no', inter=False, **kw):
    if path is None or fs is None:
        raise NASError('Bad Request')

    path = str(path)
    fs = str(fs)
    label = str(label)
    force = str(force)

    path = os.path.realpath(path)

    if path.startswith('/dev/mtd'):
        raise NASError('Invalid Device', path)

    if path.startswith('/dev/sd'):
        if len(path) == len('/dev/sda') and force != 'yes':
            raise NASError('Entire Device', path)
    elif path.startswith('/dev/md'):
        pass
    else:
        raise NASError('Invalid Device', path)

    if fs not in MKFS:
        raise NASError('Invalid File System', fs)

    mountpoints = _getmountpoints()
    if path in mountpoints:
        raise NASError('Mounted Partition')

    #device = None
    #if path.startswith('/dev/md') and 'p' not in path:
    #    device = parted.Device(path)
    #elif path == path.rstrip('0123456789'):
    #    device = parted.Device(path)

    #if device is not None:
    #    if device.busy:
    #        return {'error': 'device %s is busy at present' % device.path}

    print >> sys.stderr, '[debug]', 'try lock ...'
    lockfile = '/var/run/%s.lock' % os.path.split(path)[1]
    lockfd = os.open(lockfile, os.O_RDWR | os.O_CREAT)
    try:
        fcntl.lockf(lockfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError as error:
        if error.errno in [errno.EACCES, errno.EAGAIN]:
            os.close(lockfd)
            raise NASError('Operation Busy')

    os.ftruncate(lockfd, 0)
    os.write(lockfd, '%s\n' % path)
    os.write(lockfd, '%s\n' % fs)
    os.write(lockfd, '%s\n' % label)
    os.write(lockfd, '%s\n' % force)
    os.write(lockfd, '%s\n' % time.strftime('%Y/%m/%d %H:%M:%S %Z'))

    if fs == 'swap':
        _fixflag(path, parted.PARTITION_SWAP, 'linux-swap(v1)')
    else:
        _fixflag(path, None, fs)

    command = MKFS[fs]
    if label:
        command += ' %s %s' % (MKFSOPT[fs], label)
    command += ' %s' % path

    fcntl.lockf(lockfd, fcntl.LOCK_UN)
    os.close(lockfd)

    if not inter:
        if os.fork() > 0:
            return {'mkfs': 'path=%s, fs=%s, label=%s' % (path, fs, label)}

    lockfp = open(lockfile, 'a+')
    fcntl.lockf(lockfp, fcntl.LOCK_EX)

    (error, result) = POPEN(command, outputs=lockfp, errors=lockfp)
    lockfp.write('\nretcode: %s' % error)
    lockfp.flush()
    lockfp.seek(0, 0)
    status = echo(lockfp.read())
    fcntl.lockf(lockfp, fcntl.LOCK_UN)
    lockfp.close()
    if not inter:
        os._exit(0)
    else:
        return (error, status)

LEVELS = ['linear', 'raid0', '0', 'stripe', 'raid1', '1', 'mirror', 'raid4', '4', 'raid5', '5', 'raid6', '6', 'raid10', '10']
def mdcreate(device=None, level=None, disks=None, spare=0, metadata=None, force='no', clean='no', **kw):
    if device is None:
        raise NASError('Bad Request', 'no device')

    if level is None:
        raise NASError('Bad Request', 'no level')

    if disks is None:
        raise NASError('Bad Request', 'no disks')

    if not isinstance(disks, list):
        raise NASError('Bad Request', 'no enough disks')

    if device.startswith('/dev/md') and 'p' not in device:
        pass
    elif '/' not in device and device.islower():
        pass
    else:
        raise NASError('Invalid Raid Name', device)

    level = str(level)
    if level not in LEVELS:
        raise NASError('Invalid Raid Level', level)

    disks = map(lambda x: str(x), disks)

    for disk in disks:
        if not disk.startswith('/dev/sd'):
            raise NASError('Invalid Raid Disks', ', '.join(disks))

    if not str(spare).isdigit():
        spare = 0
    else:
        spare = int(spare)

    force = str(force)
    if force != 'yes':
        for disk in disks:
            argv = ['--examine', disk]
            (error, status) = _mdadm(argv)
            if error == 0:
                raise NASError('Invalid Disks', status)
    else:
        for disk in disks:
            argv = ['--zero-superblock', disk]
            _mdadm(argv)

    for disk in disks:
        _fixflag(disk, parted.PARTITION_RAID)

    for number in range(1, 127):
        if not os.path.lexists('/dev/md%d' % number):
            device = '/dev/md%d' % number
            break

    argv = ['--create', device, '--run', '--level=%s' % level, '--assume-clean']
    argv.append('--raid-devices=%d' % (len(disks) - spare))
    if spare:
        argv.append('--spare-devices=%d' % spare)

#    if clean == 'yes':
#        argv.append('--assume-clean')

    argv.extend(disks)

    (error, status) = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    mdpath = _getmdpath(disks[0], True)
    _mdadmconf(mdpath)
    return {'raid': 'create device'}

#def mddetail(device, **kw):
#    argv = ['--misc', '--detail', device]
#    return _mdadm(argv)

def _mdadmconf(mdpath=None):
    if mdpath is not None:
        argument = 'mdadm --misc --detail %s' % str(mdpath)
        (error, result) = POPEN(argument)
        if error != 0:
            result.close()
            mdpath = None
        while True:
            line = result.readline()
            if not line:
                mdpath = None
            line = line.strip()
            if line.startswith('UUID'):
                uuid = line.partition(':')[2].strip().split()[0]
                record = 'ARRAY %s UUID=%s' % (mdpath, uuid)
                break
        result.close()

    if mdpath is None:
        argv = ['--examine', '--scan']
        (error, record) = _mdadm(argv)
        if error != 0:
            raise NASError('System Error', record)

    if not record:
        return False


    backup(MDADMCONF)
    try:
        os.stat(MDADMCONF)
    except os.error:
        src = None
    else:
        src = open(MDADMCONF, 'r')

    dst = os.open('%s.lock' % MDADMCONF, FLAGS, 0644)
    fcntl.lockf(dst, fcntl.LOCK_EX)
    error = 0
    while error == 0:
        if src is None:
            break
        line = src.readline()
        if not line:
            break
        if line[:1] == '#':
            continue
        if record and line[:5] == 'ARRAY' and _sameraid(record.split()[1], line.split()[1]):
            error |= write(dst, '%s\n' % record.strip())
            record = ''
        elif line[:5] == 'ARRAY':
            error |= write(dst, line)

    if record and error == 0:
        error |= write(dst, '\n%s\n' % record.strip())
    os.fsync(dst)
    fcntl.lockf(dst, fcntl.LOCK_UN)
    os.close(dst)
    if src is not None:
        src.close()

    if error:
        raise NASError('write data error', MDADMCONF)

    os.rename('%s.lock' % MDADMCONF, MDADMCONF)
    unbackup(MDADMCONF)

    return True

def _sameraid(a, b):
    if not a.startswith('/dev/md'):
        return False
    return os.path.realpath(a) == os.path.realpath(b)

def mddestroy(device=None, **kw):
    if device is None:
        raise NASError('Bad Request')

    device = os.path.realpath(str(device))
    _checkmounted(device)
    backup(MDADMCONF)
    try:
        fd = open(MDADMCONF, 'w')
    finally:
        fd.close()
    try:
        os.stat(MDADMCONF)
    except os.error:
        src = None
    else:
        src = open(MDADMCONF, 'r')

    dst = os.open('%s.lock' % MDADMCONF, FLAGS, 0644)
    fcntl.lockf(dst, fcntl.LOCK_EX)
    error = 0
    while error == 0:
        if src is None:
            break
        line = src.readline()
        if not line:
            break
        if line[:1] == '#':
            continue
        if line[:5] == 'ARRAY' and _sameraid(device, line.split()[1]):
            # remove record
            continue
        else:
            error |= write(dst, line)
    os.fsync(dst)
    fcntl.lockf(dst, fcntl.LOCK_UN)
    os.close(dst)
    if src is not None:
        src.close()

    if error:
        raise NASError('write data error', MDADMCONF)

    os.rename('%s.lock' % MDADMCONF, MDADMCONF)
    unbackup(MDADMCONF)

    disks = _getmdstat(device, state=False)[2]
    if disks is None:
        raise NASError('No device %s' % device)
    argv = ['--stop', device]
    (error, status) = _mdadm(argv)
    if error == 0 and disks:
        for disk in disks:
            argv = ['--zero-superblock', disk]
            _mdadm(argv)
    return {'raid': 'destroy device'}

def mdactive(device=None, uuid=None, **kw):
    if device is not None and uuid is not None:
        device = str(device)
        uuid = str(uuid)
        argv = ['--assemble', device, '--uuid=%s' % uuid, '--run']
    else:
        argv = ['--assemble', '--scan']
    (error, status) = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    _mdadmconf()
    return {'raid': 'active device'}

def mddeactive(device=None, **kw):
    if device is None:
        raise NASError('Bad Request')
    argv = ['--stop', str(device)]
    (error, status) = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    _mdadmconf()
    return {'raid': 'deactive device'}

def mdremove(device=None, path=None, **kw):
    if device is None or path is None:
        raise NASError('Bad Request')
    argv = ['--fail', device, path]
    error, status = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    time.sleep(3)
    argv = ['--remove', device, path]
    error, status = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    return {'raid': 'remove %s' % path}

def _getsize(device):
    (state, level, devices, sync) = _getmdstat(device)
    if level != 'raid1':
        raise NASError('Only Support RAID1')
    raid = parted.Device(device)
    pathsize = raid.length
    for x in devices:
        if ' ' in x: # spare or faulty
            continue
        try:
            path = parted.Device(x)
        except:
            continue
        pathsize = path.length
        break
    return pathsize

def _check_partition(path, length, retry=3):
    try:
        device = parted.Device(path)
    except:
        raise NASError('%s Error' % path)

    if device.path[-1:].isdigit():
        _fixflag(device.path, parted.PARTITION_RAID)
        return device.path

    path = device.path
    sector = device.biosGeometry[2]
    print 'path', device.path
    try:
        deleteallpartition(device.path, inter=True)
    except CommitError:
        time.sleep(3)
        newretry = retry - 1
        if newretry < 0:
            raise
        return _check_partition(path, length, newretry)

    pathname = None
    argv = {'path': path,
            'start': sector,
            'length': length,
            'type': 'primary',
            'flag': 'raid',
            'inter': True}
    pathname, geometry = addpartition(**argv)
    time.sleep(1)
    print >> sys.stderr, pathname, os.path.exists(pathname)

    return pathname

def mdadd(device=None, path=None, **kw):
    # todo: support disk
    if device is None or path is None:
        raise NASError('Bad Request')
    pathsize = _getsize(device)
    pathname = _check_partition(path, pathsize)
    argv = ['--remove', device, 'failed']
    error, status = _mdadm(argv)
    time.sleep(3)
    argv = ['--remove', device, 'detached']
    error, status = _mdadm(argv)
    time.sleep(3)
    argv = ['--add', device, pathname]
    error, status = _mdadm(argv)
    if error != 0:
        raise NASError('System Error', status)
    return {'raid': 'add %s' % pathname}

def _mdadm(argv):
    # argv = map(str, argv)
    argument = 'mdadm --verbose %s' % ' '.join(argv)
    (error, result) = POPEN(argument)
    status = result.read()
    result.close()
    return (error, status)

def _blkid(partition, tag):
    (error, result) = POPEN('blkid -s %s %s' % (tag, partition))
    status = result.read().rstrip()
    result.close()
    if tag in status:
        return status.split('%s=' % tag)[1].replace('"', '')
    return ''

def _getfstype(partition):
    return _blkid(partition, 'TYPE')

def _getlabel(partition):
    return _blkid(partition, 'LABEL')

def _getswap():
    swaps = []
    argv = ['swapon', '-s']
    nil, status = POPEN(argv)
    while True:
        line = status.readline()
        if not line:
            break
        if not line.startswith('/'):
            continue
        path = line.split()[0]
        swaps.append(path.split('\\')[0])
    status.close()
    return swaps

def allinone(level='1', fs='ext4', isinternal='yes', disks=None, **kw):
    print >> sys.stderr, '[wizard]', 'try lock ...'

    lockfile = '/var/run/allinone.lock'
    lockfd = os.open(lockfile, os.O_RDWR | os.O_CREAT)
    try:
        fcntl.lockf(lockfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError as error:
        if error.errno in [errno.EACCES, errno.EAGAIN]:
            lockfp = os.fdopen(lockfd)
            string = lockfp.read()
            lockfp.close()
            raise NASError('Operation Busy', string)
    _allinone(level, fs, lockfd, isinternal, disks)
    fcntl.lockf(lockfd, fcntl.LOCK_UN)
    os.close(lockfd)
    return {'allinone': 'done'}

def partition(path=None, **kw):
    if path is None:
        raise NASError('Bad Request')
    _checkdevice(path)
    _partition(path)
    return {'partition': 'done'}

def _partition(disk, retry=3, need_swap=True):
    cmd = ['dd', 'if=/dev/zero', 'of=%s' % disk, 'bs=1M', 'count=10']
    code, pipe = POPEN(cmd)
    if code != 0:
        raise NASError("Error erasing partitions on %s" % disk)

    try:
        device = parted.Device(disk)
    except parted.DeviceException:
        raise NASError('Wizzard Error', 'Cannot Parted')

    cylinder, head, sector = device.biosGeometry
    swaplength = 512 * 1024 * 1024 / device.sectorSize
    if swaplength > device.length:
        raise NASError('Wizzard Error', 'Internal Disk is too small')
    swapstart = (device.length - swaplength) / sector * sector

    home = None
    swap = None

    homeargv = {'path': disk,
                'start': sector,
                'type': 'primary',
                'inter': True}
    if need_swap:
        homeargv['length'] = (swapstart - sector - 1)
    home, geometry = addpartition(**homeargv)
    time.sleep(1)
    print >> sys.stderr, 'home', home, os.path.exists(home)

    if need_swap:
        swapargs = {'path': disk,
                    'start': (geometry.end + 1),
                    'type': 'primary',
                    'inter': True}
        swap, nil = addpartition(**swapargs)
        print >> sys.stderr, 'swap', swap, os.path.exists(swap)
        time.sleep(1)
        error, status = mkfs(swap, 'swap', 'SWAP', 'yes', inter=True)
        print >> sys.stderr, error, status
        time.sleep(1)

    return home, swap

def _isflash(path):
    try:
        stat = os.lstat(path)
    except OSError:
        return True
    dev_path = "/sys/dev/block/%s:%s" % (os.major(stat.st_rdev), os.minor(stat.st_rdev))
    real_path = os.readlink(dev_path)
    if "host4" in real_path:
        return True
    return False


def _get_internal_disks():
    internal = []
    for path in glob.glob('/dev/[h|s]d[a-z]'):
        try:
#            if _isflash(path):
#                continue
            device = parted.Device(path)
        except parted.DeviceException:
            continue
        except parted.IOException:
            continue
        # should use udevadm to get type ...
        if device.model.startswith('ATA '):
            internal.append(device.path)

    internal.sort()
    return internal

def _has_home_mounted():
    hashome = False
    for dev, mountpoints in _getmountpoints().items():
        for mountpoint in mountpoints:
            if mountpoint == HOME:
                hashome = True
                break
    return hashome

def _allinone(level, fs, lockfd, isinternal='yes', disks=None):
    if not disks:
        disks = _get_internal_disks()
        if not disks:
            raise NASError('Invalid Request', 'No Internal Disks')
        isinternal = True
    else:
        disks = [ d.strip() for d in disks.split(',') ]
        isinternal = (isinternal.lower() == 'yes' and True or False)

    if len(disks) > 1 and level not in LEVELS:
        raise NASError('RAID Level Error', level)

    print >> sys.stderr, '[wizzard] determing raid'
    swaps = _getswap()
    mdpaths = []
    for disk in disks:
        mdpath = _getmdpath(disk, False)
        if mdpath:
            mdpaths.append(mdpath)
        for p in swaps:
            if p.startswith(disk):
                nil, status = POPEN(['swapoff', p])
                status.close()

    need_swap = False
    if isinternal:
        need_swap = True
        swaps = _getswap()
        other_disks = set(_get_internal_disks()) - set(disks)
        for swap, disk in itertools.product(swaps, other_disks):
            if swap.startswith(disk):
                need_swap = False
                break

    if isinternal and os.path.exists('/proc/mdstat'):
        raid = True
    else:
        raid = False

    # umount normal device and raid devices (if there is any)
    print >> sys.stderr, '[wizzard] umount devices'
    for device in itertools.chain(disks, mdpaths):
        mountpoints = _getmountpoints()
        for m in mountpoints:
            if m.startswith(device):
                umount(mountpoints[m], 'yes')

    # Destroy raid if there is any
    if isinternal and mdpaths:
        for mdpath in set(mdpaths):
            mddestroy(mdpath)

    # delete partitions
    print >> sys.stderr, '[wizzard] delete partitions'
    for device in disks:
        # delete zero superblock, if possible
        argv = ['--zero-superblock', device]
        _mdadm(argv)
        deleteallpartition(device)

    # Create partitions (all partitions are removed in this step)
    raiddevices = []
    for disk in disks:
        home, swap = _partition(disk, need_swap=need_swap)
        if swap:
            mount(swap, 'swap', 'yes')
#            break
        raiddevices.append(home)

    # Create soft raid if necessary
    if len(raiddevices) == 1 or not raid:
        path = raiddevices[0]
        homepath = path
    else:
        level = str(level)
        for f in range(1, 127):
            if not os.path.lexists('/dev/md%d' % f):
                break
        mdcreate(device='/dev/md%d' % f, level=level, disks=raiddevices, force='yes')
        mdpath = _getmdpath(raiddevices[0], True)
        homepath = mdpath

    # Make file system
    error, status = mkfs(homepath, fs, 'HOME', 'yes', inter=True)
    print >> sys.stderr, error, status
    nil, status = POPEN('umount %s' % homepath)
    status.close()
    time.sleep(1)

    # Mount partitions
    if isinternal and not _has_home_mounted():
        mount(homepath, 'home', 'yes')
    else:
        static = isinternal and 'yes' or 'no'
        mount_point = os.path.basename(homepath)
        mount(homepath, mount_point, static)

    return {'allinone': 'done'}

def _getmdpath(disk, strict=False):
    mdpath = None
    file = open('/proc/mdstat')
    while True:
        line = file.readline()
        if not line:
            break
        if not line.startswith('md'):
            continue
        if strict:
            raiddevice = '%s[' % os.path.split(disk)[1]
        else:
            raiddevice = os.path.split(disk)[1]
        if raiddevice in line:
            mdpath = os.path.join('/dev', line.partition(':')[0].strip())
            break
    file.close()
    return mdpath

def _getfree(directory):
    vfs = os.statvfs(directory)
    return 100 * vfs.f_bavail / vfs.f_blocks

def checkpartitions():
    disks = []
    for path in glob.glob('/dev/[h|s]d[a-z]'):
        error, pipe = POPEN(['hdparm', '-C', path])
        if error != 0:
            pipe.close()
            continue

        status = pipe.read().strip()
        pipe.close()

        if 'standby' in status:
            disks.append(path)

    storage = []
    mountpoints = _getmountpoints()
    for (name, value) in mountpoints.iteritems():
        if name.startswith('/dev/sd'):
            if name in disks:
                continue
            free = _getfree(value[0])
            if free < 5:
                storage.append((name, value[0], free))

    if storage:
        subject = '[NAS] Storage Capacity Less than 5%'
        body =[]
        for name, value, free in storage:
            body.append('device=%s, mountpoint=%s, free=%s' % (name, value, free))
        body.append('For more information, please visit your NAS %s' % os.uname()[1])
        _notify(subject, '\r\n'.join(body))


    # print storage

# vim: set sta sw=4 et:
